/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_RAW_SOCKET_HH
#define SENDS_RAW_SOCKET_HH

#include "socket_api.hh"

namespace sends {

    /**  The default socket api stuff.
      *  \brief Socket API.
      *  \author J. Zweizig
      *  \version 1.0; modified April 24, 2008
      */
    class raw_socket : public socket_api {
    public:
	/**  Raw socket constructor
	  *  \brief Constructor
	  */
	raw_socket(void);

	/**  Disconnect and close the socket. Destroy the socket object.
	  *  \brief Destructor.
	  */
	~raw_socket(void);

	/**  Accept a requested connection and create a client socket. 
	  *  \brief Accept a connection request.
	  *  \return Client socket pointer if accept succedes, NULL otherwise.
	  */
        raw_socket* accept(void);

	/**  Bind the socket to the specified IP and port number.
	  *  \brief Bind the socket to a port.
	  *  \param port Port name string in the form "<ip-addr>:<port>"
	  *  \return Socket error code
	  */
	sock_err bind(const std::string& port);

	/**  Disconnect and close the socket.
	  *  \brief Close the socket.
	  *  \return Socket error code
	  */
	sock_err close(void);

	/**  Connect the socket to the specified peer IP and port number.
	  *  \brief Connect the socket to a peer.
	  *  \param peer Port name string in the form "<ip-addr>:<port>"
	  *  \return Socket error code
	  */
	sock_err connect(const std::string& peer);

	/**  Listen to the specified port address, Queue the specified 
	  *  number of connection requests pending acceptance.
	  *  \brief Listen for connection requests.
	  *  \param port Local port name to listen to.
	  *  \param lq   Length of connection request queue
	  *  \return Socket error code.
	  */
	sock_err listen(const std::string& port, int lq);

	/**  Read up to the specified number of bytes from the socket into 
	  *  the buffer.
	  *  \brief Read available data.
	  *  \param buf  Pointer to buffer to receive data.
	  *  \param len Maximum length of data to be read (in bytes).
	  *  \return Number of data bytes read.
	  */
	virtual long read_available(char* buf, long len);

	/**  Get a socket id number. In general this will be the system fd
	  *  for the socket.
	  *  \brief Get the socket id.
	  *  \return Socket id for use in select.
	  */
	 virtual int socket_id(void) const;

	/**  Wait at most the specified number of seconds for data
	  *  to arrive for reading.
	  *  \brief Wait for data.
	  *  \param nsec Maximum number of seconds to wait for data
	  *  \return Integer with the following values:
	  *      - -1 Systerm error occurred (e.g. EINTR)
	  *      -  0 Request timed out.
	  *      -  1 data available on port.
	  */
	int wait_data(double nsec);

	/**  Write the specified number of bytes into the buffer.
	  *  \brief Write data.
	  *  \param buf  Pointer to data to be written
	  *  \param len Length of data to be written (in bytes).
	  *  \return Number of data bytes written or -1 on error.
	  */
	virtual long write(const char* buf, long len);

    protected:
	/**  Construct a raw_socket from an existing unix socket.
	  *  \brief construct a socket
	  *  \param fd File descriptor for existing socket.
	  */
	raw_socket(int fd);

    protected:
	/** Opened socket file number or -1.
	 */
	int mSocket;
    };


    //==================================  Inline methods.
    inline int
    raw_socket::socket_id(void) const {
	return mSocket;
    }

}  // namespace sends

#endif // !defined(SENDS_RAW_SOCKET_HH)
