/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "src_update.hh"
#include "frame_source.hh"
#include "gpstime.hh"
#include "EggTimer.hh"
#include "Interval.hh"
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <stdexcept>

using namespace std;
using namespace sends;

//======================================  src_update constructor
src_update::src_update(void) 
    : mCurrentTime(0), mDebug(0), mMsgState(kNormal), 
      mSumLate(0.0), mNMeasure(0)
{
}

//======================================  src_update destructor
src_update::~src_update(void) {
}

//======================================  test a file
//    Feb 23, 2009: Tested/working.
inline bool 
src_update::test_file(const frame_loc& fl, gps_type gps) const {
    string path = fl.file_path(gps, false);
    bool rc = (access(path.c_str(), R_OK) == 0);
    if (mDebug > 2) cout << "test_file(" << name() << ", " << gps 
			 << ") = " << rc << endl;
    return rc;
}

//======================================  Set the update mode
gps_type 
src_update::beg_time(const loc_list& ll, gps_type lo, gps_type hi) const {
    ll.read_lock();
    const frame_loc& fl = ll.back();
    gps_type tHi = fl.align_gps(hi);
    gps_type tLo = fl.align_gps(lo);
    gps_type dt  = fl.duration();
    while (tHi > tLo + dt){
	gps_type tMid = fl.align_gps((tHi + tLo) >> 1);
	if (test_file(fl, tMid)) tHi = tMid;
	else                     tLo = tMid;
    }
    if (mDebug > 1) cout << "beg_time(" << name() << ") = " << tHi << endl;
    ll.unlock();
    return tHi;
}

//======================================  Set the update mode
gps_type 
src_update::end_time(const loc_list& ll, gps_type lo, gps_type hi) const {
    ll.read_lock();
    const frame_loc& fl = ll.back();
    gps_type tHi = fl.align_gps(hi);
    gps_type dt  = fl.duration();
    gps_type tLo = fl.align_gps(lo);
    if (mDebug > 2) cout << "end_time: scan(" << name() << "; " << tLo 
			 << ":" << tHi <<":" << dt << ")" << endl;
    while (tHi > tLo + dt) {
	gps_type tMid = fl.align_gps((tHi + tLo) >> 1);
	if (test_file(fl, tMid)) tLo = tMid;
	else                     tHi = tMid;
    }
    if (tHi != tLo && test_file(fl, tHi)) tLo = tHi;
    if (mDebug > 1) cout << "end_time(" << name() << ") = " << tLo << endl;
    ll.unlock();
    return tLo;
}

//======================================  Return descriptive name
string
src_update::name(void) const {
    string name_str = loc_list::arch_name(mArchive);
    name_str += ":";
    name_str += mSrcType;
    return name_str;
}

//======================================  Find the end of the data
void 
src_update::now_available(void) {
    loc_list* ll = Source_List.list_ptr(mArchive, mSrcType);
    if (!ll) throw runtime_error("src_update: couldn't find frame_loc");
    gps_type now  = gps_now(); 

    //----------------------------------  Look for start time if not valid
    if (!mCurrentTime) {
	mCurrentTime = ll->back().start();
	if (!mCurrentTime) {
	    mCurrentTime = beg_time(*ll, 0, now);
	    ll->write_lock();
	    ll->back().set_start(mCurrentTime);
	    ll->unlock();
	}
	if (mDebug) {
	    cout << "src_update: source " << name() << " start time set to "
		 << mCurrentTime << endl;
	}
    }

    //----------------------------------  Look for start time if not valid
    update(*ll, end_time(*ll, mCurrentTime, now));
}

//======================================  Set the debug print level.
void
src_update::set_debug(int dbg) {
    mDebug = dbg;
}

//======================================  Set the update mode
void
src_update::set_scan_mode(double latency, double retry) {
    mSchedMode = kScan;
    mStep      = retry;
    mLatency   = latency;
}

//======================================  Set the update mode
void
src_update::set_sched_mode(double latency, double retry) {
    mSchedMode = kSched;
    mStep      = retry;
    mLatency   = latency;
}

//======================================  Set the update mode
void
src_update::set_synch_mode(const std::string& file, double late, double step) {
    mSchedMode = kSynch;
    mSynchFile = file;
    mStep      = step;
    mLatency   = late;
}

//======================================  Set the source class
void
src_update::set_source(loc_list::arch_type a, const string& s, gps_type t) {
    mArchive     = a;
    mSrcType     = s;
    mCurrentTime = t;
}

//======================================  Start the thread
int
src_update::start(void) {

    //----------------------------------  Log the thread.
    cout << "Starting updater for " << name() << " mode: ";
    switch (mSchedMode) {
    case kSynch: 
	cout << "synch file: " << mSynchFile;
	break;
    case kSched:
	cout << "sched";
	break;
    case kScan:
	cout << "scan";
	break;
    default:
	cout << "unknown";
    }
    cout << " delay: " << mLatency << " step: " << mStep << endl;

    //----------------------------------  Start the thread
    int rc = 0;
    try {
	now_available();
	rc = start_thread();
    } catch (runtime_error& e) {
	cout << "Exception caught starting update thread: " << e.what() << endl;
	cout << "Update failed for " << name() << endl;
	rc = 1;
    }
    return rc;
}

//======================================  Stop the thread
int
src_update::stop(void) {
    stop_thread();
    mSynchEvent.lock();
    mSynchEvent.release();
    mSynchEvent.unlock();
    return 0;
}

//======================================  Thread execution
void*
src_update::thread_entry(void) {
    while (run()) {
	if (mDebug > 2) {
	    cout << "src_update: " << name() <<" iteration" << endl;
	}

	double    tNow  = gps_now_d(); 
	double    tWake = tNow + mStep;
	loc_list* ll    = Source_List.list_ptr(mArchive, mSrcType);
	if (!ll) throw runtime_error("src_update: couldn't find loc_list");

	switch (mSchedMode) {
	//------------------------------  Synchronize to a file.
	case kSynch: {
	    gps_type tEnd  = 0;
	    gps_type dt    = ll->back().duration();
	    if (!mStep) tWake += dt;
	    ifstream inf(mSynchFile.c_str(), ios::in);
	    if (!inf.is_open()) {
		if (mMsgState == kNormal) {
		    cout << "Unable to open synch file: "<< mSynchFile << endl;
		    mMsgState = kNoSync;
		}
	    } else {
		inf >> tEnd;
		inf.close();
		if (tEnd > mCurrentTime) {
		    tWake = tEnd + dt + mLatency;
		}
		if (mMsgState != kNormal) {
		    cout << "Synch file OK: "<< mSynchFile << endl;
		    mMsgState = kNormal;
		}
		update(*ll, tEnd + dt);
	    }
	    break;
	}

	//------------------------------  Schedule search
	case kSched: {
	    double step  = mStep;
	    if (!mStep) {
		step = ll->back().duration();
		tWake += step;
	    }
	    gps_type tEnd  = end_time(*ll, mCurrentTime, gps_type(tNow));
	    tWake = long((tWake - mLatency) / step) * step + mLatency;
	    update(*ll, tEnd);
	    break;
	}

	//------------------------------  Schedule rescan of directory.
	case kScan: {
	    EggTimer et;
	    Source_List.rescan(mArchive, mSrcType);
	    cout << "Rescan complete... Elapsed time = " << double(et.elapsed())
		 << endl;
	    tWake = tNow + mLatency;
	    gps_type tEnd =  ll->back().stop();
	    update(*ll, tEnd);
	    break;
	}

	//------------------------------  Unknown update technique
	default:
	    tWake = 0;
	}

	//------------------------------  schedule a new update
	if (tWake == 0.0) break;
	double tWait = tWake - gps_now_d();
	mSynchEvent.lock();
	mSynchEvent.wait_timed(tWait);
	mSynchEvent.unlock();
    }
    return 0;
}

//======================================  Update the current time.
void 
src_update::update(loc_list& ll, gps_type t) {
    if (t <= mCurrentTime) return;
    ll.write_lock();
    frame_loc& endfl = ll.back();
    if (t > endfl.stop()) endfl.set_stop(t);
    mCurrentTime = t;
    if (mDebug) {
	cout << "src_update: source " << name() << " end time set to "
	     << mCurrentTime << endl;
    }
    ll.unlock();
    mWaitGateway.lock();
    mWaitGateway.release();
    mWaitGateway.unlock();
}

//======================================  Wait a maximum time for new data
bool 
src_update::timed_wait(double tmax) const {
    bool rc = false;
    thread::semlock lock(mWaitGateway);
    if (tmax < 0) {
	mWaitGateway.wait();
	rc = true;
    } else {
	rc = mWaitGateway.wait_timed(tmax);
    }
    return rc;
}

//======================================  Wait for new data
void
src_update::wait(void) const {
    timed_wait(-1.0);
}

//======================================  Wait for data from a specified time
bool 
src_update::wait(gps_type t, double dt) const {
    if (mDebug) cout << "src_update: wait " << dt << "s for update on " 
		     << name() << " current= " << mCurrentTime << " waitfor= "
		     << t << endl;
    while (t >= mCurrentTime) {
	if (!timed_wait(dt)) {
             if (mDebug) cout << "src_update: wait " << name() << " timed out."
			      << endl;
 	     return false;
	}
	if (mDebug) cout << "src_update: wake up " << name() << " at "
			 << mCurrentTime << endl;
    }
    if (mDebug) cout << "src_update: " << name() << " wait complete." << endl;    
    return true;
}
