/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "sslbase.hh"
#include "openssl/bio.h"
#include "openssl/ssl.h"
#include "openssl/err.h"
#include <stdexcept>
#include <iostream>

using namespace sends;
using namespace std;

extern "C" {

    static int 
    password_callback(char *buf, int size, int rwflag, void *userdata) {
	strncpy(buf, (const char*)userdata, size);
	return strlen(buf);
    }

}

//======================================  Constructor
ssl_ctx::ssl_ctx(void) 
    : mCtx(0)
{
    //----------------------------------  Initialize openssl
    SSL_library_init();
    SSL_load_error_strings();
    ERR_load_BIO_strings();
    OpenSSL_add_all_algorithms();

    //----------------------------------  Initialize openssl
    mCtx = SSL_CTX_new(SSLv23_server_method());
    if (!mCtx) error("constructing context");
}

//======================================  Destructor
ssl_ctx::~ssl_ctx(void) {
    if (mCtx) SSL_CTX_free(mCtx);
    mCtx = 0;
}

//======================================  Generate a bio.
bio_st*
ssl_ctx::new_ssl(void) {
    return BIO_new_ssl(mCtx, 0);
}

//======================================  Make a secure connection.
void
ssl_ctx::secure(const string& certpath, const string& keypath, 
		const string& pass) {
    SSL_CTX_set_default_passwd_cb(mCtx, password_callback);
    SSL_CTX_set_default_passwd_cb_userdata(mCtx, (void*)pass.c_str());

    int err = SSL_CTX_use_certificate_file(mCtx, certpath.c_str(),
					   SSL_FILETYPE_PEM);
    if (err < 1) error("reading certificate");

    err = SSL_CTX_use_PrivateKey_file(mCtx, keypath.c_str(),
				      SSL_FILETYPE_PEM);
    if (err < 1) error("reading private key");
}

//======================================  Throw exception with last error code.
void
ssl_ctx::error(const string& msg) {
    unsigned long err = ERR_get_error();
    char errstr[128];
    ERR_error_string_n(err, errstr, sizeof(errstr));
    throw runtime_error(string("While ") + msg + ": " + errstr);
}

//======================================  Default constructor
ssl_bio::ssl_bio(void) 
    : mBio(0)
{
}

//======================================  Default constructor
ssl_bio::ssl_bio(bio_st* bio) 
    : mBio(bio)
{
}

//======================================  Context constructor
ssl_bio::ssl_bio(ssl_ctx& ctx) 
    : mBio(0)
{
    reset(ctx.new_ssl());
    set_ssl_mode(SSL_MODE_AUTO_RETRY);
}

//======================================  Copy constructor
//ssl_bio::ssl_bio(ssl_bio& bio)
//    : mBio(bio.release())
//{}

//======================================  Destructor
ssl_bio::~ssl_bio(void) {
    close();
}


//======================================  Assignment operator
ssl_bio&
ssl_bio::operator=(ssl_bio& bio) {
    reset(bio.release());
    return *this;
}

//======================================  Set the SSL mode flags
void
ssl_bio::set_ssl_mode(long mode) {
    SSL* ssl = 0;
    BIO_get_ssl(mBio, &ssl);
    SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
}

//======================================  Get the socket ID.
int
ssl_bio::socket_id(void) const {
    return BIO_get_fd(mBio, 0);
}

//======================================  Accept a request.
ssl_bio*
ssl_bio::accept(void) {
    if (BIO_do_accept(mBio) <= 0) ssl_ctx::error("accepting connection");
    bio_st* bio = BIO_pop(mBio);
    return new ssl_bio(bio);
}

//======================================  Bind the socet to an local port
ssl_bio::sock_err
ssl_bio::bind(const std::string& port) {
    return serr_OK;
}

//======================================  Free the BIO.
ssl_bio::sock_err
ssl_bio::close(void) {
    if (mBio) {
	BIO_ssl_shutdown(mBio);
	BIO_free_all(mBio);
    }
    mBio = 0;
    return serr_OK;
}

//======================================  Free the BIO.
ssl_bio::sock_err
ssl_bio::connect(const std::string& peer) {
    return serr_OK;
}

//======================================  Perform authentication.
ssl_bio::sock_err
ssl_bio::handshake(void) {
    sock_err rc = serr_OK;
    if (BIO_do_handshake(mBio) <= 0) {
	if (verbose()) ssl_ctx::error("doing handshake");
	rc = serr_failure;
    }
    return rc;
}

//======================================  Listen on the specified port
ssl_bio::sock_err
ssl_bio::listen(const std::string& port, int lq) {
    sock_err rc = serr_OK;

    /* Test BIO Accept */ 
    BIO* acc = BIO_new_accept(const_cast<char*>(port.c_str()));
    if (!acc) {
	ssl_ctx::error("creating server socket");
	return serr_failure;
    }

    if (BIO_set_accept_bios(acc, release()) <= 0) {
	return serr_failure;
    }

    reset(acc);
    if (BIO_do_accept(mBio) <= 0) {
	if (verbose()) ssl_ctx::error("binding server socket");
	rc = serr_failure;
    }
    return rc;
}

//======================================  Read a buffer
long
ssl_bio::read_available(char* buf, long maxlen) {
    int len = BIO_read(mBio, buf, maxlen);
    if (len < 1) set_error(serr_failure);
    return len;
}

//======================================  Release the bio object.
bio_st*
ssl_bio::release(void) {
    bio_st* rc = mBio;
    mBio = 0;
    return rc;
}

//======================================  Accept a request.
void
ssl_bio::reset(bio_st* bio) {
    if (bio != mBio) {
	close();
	mBio = bio;
    }
}

//======================================  Read a buffer
long
ssl_bio::write(const char* buf, long maxlen) {
    long nB = 0;
    while (nB < maxlen) {
	int len = BIO_read(mBio, const_cast<char*>(buf), maxlen);
	if (len < 1) {
	    set_error(serr_failure);
	    nB = -1;
	    break;
	} else {
	    nB += len;
	}
    }
    return nB;
}
